from __future__ import annotations

from typing import Any, Iterator, Tuple


class NameSpace:
    """A namespace class that allows attribute access dynamically through item access."""

    def __init__(self, mapping: dict = None, /, **kwargs: Any) -> None:
        self(mapping, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self])})"

    def __call__(self, mapping: dict = None, /, **kwargs: Any) -> NameSpace:
        for name, _ in self:
            del self[name]

        if mapping is not None:
            vars(self).update(mapping)

        vars(self).update(kwargs)

        return self

    def __len__(self) -> int:
        return len([item for item in self])

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name if name is not None else "__none__")

    def __setitem__(self, name: str, val: Any) -> None:
        setattr(self, name if val is not None else "__none__", val)

    def __delitem__(self, name: str) -> None:
        delattr(self, name if name is not None else "__none__")

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        return iter([(name, val) for name, val in vars(self).items() if not name.startswith("_")])

    def __contains__(self, other: Any) -> bool:
        return other in set(vars(self).keys())



from __future__ import annotations

import datetime as dt
from typing import Any, Union

from django.utils.functional import cached_property
from dateutil.relativedelta import relativedelta
import parsedatetime

from .str import Str
from .enums import Enum, ValueEnum


class FormatCode(ValueEnum):
    """An Enum containing all the formatcodes used by DateTime.strptime() and DateTime.strftime(), subdivided into further Enums."""
    class TIMEZONE(ValueEnum):
        NAME = "%Z"

    class WEEKDAY(ValueEnum):
        SHORT, FULL, NUM = "%a", "%A", "%w"

    class WEEK(ValueEnum):
        OF_YEAR_STARTING_MONDAY, OF_YEAR_STARTING_SUNDAY = "%W", "%U"

    class YEAR(ValueEnum):
        WITHOUT_CENTURY, WITH_CENTURY = "%y", "%Y"

    class MONTH(ValueEnum):
        SHORT, FULL, NUM = "%b", "%B", "%m"

    class DAY(ValueEnum):
        NUM, OF_YEAR = "%d", "%j"

    class HOUR(ValueEnum):
        H24, H12, AM_PM = "%H", "%I", "%p"

    class MINUTE(ValueEnum):
        NUM = "%M"

    class SECOND(ValueEnum):
        NUM = "%S"

    class MICROSECOND(ValueEnum):
        NUM = "%f"


class WeekDayName(Enum):
    """An Enum holding the days of the week."""
    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY = "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"


class MonthName(Enum):
    """An Enum holding the months of the year."""
    JANUARY, FEBRUARY, MARCH, APRIL, MAY, JUNE = "january", "february", "march", "april", "may", "june"
    JULY, AUGUST, SEPTEMBER, OCTOBER, NOVEMBER, DECEMBER = "july", "august", "september", "october", "november", "december"


class Date(dt.date):
    """
    A subclass of the stdlib datetime.date class with additional useful methods.
    All normal datetime attributes are shadowed by PascalCase attributes which have properties providing various string representations of that attribute.
    E.g. Date.today().WeekDay.full, or Date.today().Year.without_century
    """
    FormatCode, WeekDayName, MonthName, calendar = FormatCode, WeekDayName, MonthName, parsedatetime.Calendar()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(self.to_isoformat())})"

    def to_stdlib(self) -> dt.date:
        """Create the equivalent datetime.date object from this DateTime."""
        return dt.date.fromordinal(self.toordinal())

    def to_ordinal(self) -> int:
        return self.toordinal()

    def to_isoformat(self) -> str:
        """Create an isoformat date string from this Date with several options."""
        return self.strftime(f"{FormatCode.YEAR.WITH_CENTURY}-{FormatCode.MONTH.NUM}-{FormatCode.DAY.NUM}")

    def to_format(self, format_string: str) -> str:
        """Create a date string from this Date using a format string."""
        return self.strftime(format_string)

    def to_filetag(self) -> str:
        """Create a string suitable to be used as a date (not datetime!) tag within a filename."""
        return self.strftime(f"{FormatCode.YEAR.WITH_CENTURY}{FormatCode.MONTH.NUM}{FormatCode.DAY.NUM}")

    def to_casual(self, full_month: bool = False, day_first: bool = True, day_suffix: bool = False) -> str:
        """Create a human-readable datetime string from this DateTime with several options."""
        day, month, year = self.Day.with_suffix if day_suffix else self.day, self.Month.full if full_month else self.Month.short, self.Year.with_century
        return " ".join([str(item) for item in ((day, month, year) if day_first else (month, day, year))])

    @classmethod
    def today(cls, **kwargs: Any) -> Date:
        """Create a DateTime from today at midnight."""
        return cls.from_date(dt.date.today())

    @classmethod
    def from_date(cls, date: dt.date) -> Date:
        """Create a DateTime from a datetime.date object."""
        return cls.fromordinal(date.toordinal())

    @classmethod
    def from_ordinal(cls, ordinal: int) -> Date:
        """Create a DateTime from a datetime.datetime object."""
        return cls.fromordinal(ordinal)

    @classmethod
    def from_isoformat(cls, date_string: str) -> Date:
        """Create a DateTime from a datetime.datetime object."""
        return cls.fromisoformat(date_string)

    @classmethod
    def from_format(cls, date_string: str, format_string: str) -> Date:
        """Create a Date from a string representing a date and a format spec."""
        return DateTime.strptime(date_string, format_string).to_date()

    @classmethod
    def from_datelike(cls, datelike: Union[dt.datetime, dt.date, int, str]) -> Date:
        """Create a DateTime from a valid python object. Raises TypeError if an unsupported type is passed. Raises ValueError if an invalid instance of a supported type (like 'str' or 'int') is passed."""
        if isinstance(datelike, dt.datetime):
            return DateTime.from_datetime(datelike).to_date()
        elif isinstance(datelike, cls):
            return datelike
        elif isinstance(datelike, dt.date):
            return cls.from_date(datelike)
        elif isinstance(datelike, int):
            return cls.from_ordinal(datelike)
        elif isinstance(datelike, str):
            try:
                return cls.from_isoformat(datelike)
            except ValueError:
                if (ret := DateTime.from_string(datelike)) is not None:
                    return ret.to_date()
                else:
                    raise ValueError(f"Could not parse stringlike value '{datelike}' to type '{cls.__name__}'.")
        else:
            raise TypeError(f"Unsupported type '{type(datelike)}' for inference to type '{cls.__name__}'.")

    @cached_property
    def WeekDay(self) -> WeekDay:
        return WeekDay(self)

    @cached_property
    def Week(self) -> Week:
        return Week(self)

    @cached_property
    def Year(self) -> Year:
        return Year(self)

    @cached_property
    def Month(self) -> Month:
        return Month(self)

    @cached_property
    def Day(self) -> Day:
        return Day(self)


class DateTime(Date, dt.datetime):
    """
    A subclass of the stdlib datetime.datetime class with additional useful methods.
    All normal datetime attributes are shadowed by PascalCase attributes which have properties providing various string representations of that attribute.
    E.g. DateTime.now().WeekDay.full, or DateTime.now().Year.without_century
    """
    nanosecond = 0

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(self.to_isoformat())})"

    def delta(self, *args: Any, **kwargs: Any) -> DateTime:
        """Add/subtract the given amount of time units (as keyword arguments) to this DateTime. E.g. DateTime.now().delta(days=-3, hours=15)"""
        return self.fromisoformat((self + relativedelta(*args, **kwargs)).isoformat())

    def truncate_time(self) -> DateTime:
        return type(self)(self.year, self.month, self.day)

    def to_date(self) -> Date:
        """Create the equivalent datetime.date object from this DateTime."""
        return Date.fromordinal(self.toordinal())

    def to_stdlib(self) -> dt.datetime:
        """Create the equivalent datetime.datetime object from this DateTime."""
        return dt.datetime.fromisoformat(self.isoformat())

    def to_isoformat(self, time: bool = True, timezone: bool = False) -> str:
        code = self.FormatCode
        text = self.to_format(f"{code.YEAR.WITH_CENTURY}-{code.MONTH.NUM}-{code.DAY.NUM}")

        if time and (self.hour or self.minute or self.second):
            text = self.to_format(f"{text} {code.HOUR.H24}:{code.MINUTE.NUM}:{code.SECOND.NUM}")

        if timezone and self.tzinfo is not None:
            text = f"{text}{self.tzinfo}"

        return text

    def to_format(self, format_string: str) -> str:
        """Create a datetime string from this DateTime using a format string."""
        return self.strftime(format_string)

    def to_logformat(self, labels: bool = False) -> str:
        """Create a precise datetime string suitable to be used for logging."""
        code = FormatCode
        if labels:
            return self.strftime(f"[{code.YEAR.WITH_CENTURY}-{code.MONTH.NUM}-{code.DAY.NUM} {code.HOUR.H24}h {code.MINUTE.NUM}m {code.SECOND.NUM}s {code.MICROSECOND.NUM}ms]")
        else:
            return self.strftime(f"[{code.YEAR.WITH_CENTURY}-{code.MONTH.NUM}-{code.DAY.NUM} {code.HOUR.H24}:{code.MINUTE.NUM}:{code.SECOND.NUM}:{code.MICROSECOND.NUM}]")

    @classmethod
    def from_datetime(cls, datetime: dt.datetime) -> DateTime:
        """Create a DateTime from a datetime.datetime object."""
        return cls.from_isoformat(datetime.isoformat())

    @classmethod
    def from_format(cls, date_string: str, format_string: str) -> DateTime:
        """Create a Date from a string representing a datetime and a format spec."""
        return DateTime.strptime(date_string, format_string)

    @classmethod
    def from_string(cls, text: str) -> DateTime:
        """Attempt to parse a string of text into a DateTime. Returns None upon failure."""
        val, code = cls.calendar.parse(text)
        return cls(*val[:6 if code == 3 else 3]) if code in (1, 3) else None

    @classmethod
    def from_datelike(cls, datelike: Union[dt.datetime, dt.date, int, str]) -> DateTime:
        """Create a DateTime from a valid python object. Raises TypeError if an unsupported type is passed. Raises ValueError if an invalid instance of a supported type (like 'str' or 'int') is passed."""
        if isinstance(datelike, cls):
            return datelike
        elif isinstance(datelike, dt.datetime):
            return cls.from_datetime(datelike)
        elif isinstance(datelike, dt.date):
            return cls.from_date(datelike)
        elif isinstance(datelike, int):
            return cls.from_ordinal(datelike)
        elif isinstance(datelike, str):
            try:
                return cls.from_isoformat(datelike)
            except ValueError:
                if (ret := cls.from_string(datelike)) is not None:
                    return ret
                else:
                    raise ValueError(f"Could not parse stringlike value '{datelike}' to type '{cls.__name__}'.")
        else:
            raise TypeError(f"Unsupported type '{type(datelike)}' for inference to type '{cls.__name__}'.")

    @cached_property
    def TimeZone(self) -> TimeZone:
        return TimeZone(self)

    @cached_property
    def Hour(self) -> Hour:
        return Hour(self)

    @cached_property
    def Minute(self) -> Minute:
        return Minute(self)

    @cached_property
    def Second(self) -> Second:
        return Second(self)

    @cached_property
    def MicroSecond(self) -> MicroSecond:
        return MicroSecond(self)


class DateAccessor:
    def __init__(self, date: Date) -> None:
        self._date = date


class TimeZone(DateAccessor):
    @property
    def name(self) -> str:
        return self._date.strftime(FormatCode.TIMEZONE.NAME)


class WeekDay(DateAccessor):
    @property
    def short(self) -> str:
        return self._date.strftime(FormatCode.WEEKDAY.SHORT)

    @property
    def full(self) -> str:
        return self._date.strftime(FormatCode.WEEKDAY.FULL)

    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.WEEKDAY.NUM)


class Week(DateAccessor):
    @property
    def of_year_starting_monday(self) -> str:
        return self._date.strftime(FormatCode.WEEK.OF_YEAR_STARTING_MONDAY)

    @property
    def of_year_starting_sunday(self) -> str:
        return self._date.strftime(FormatCode.WEEK.OF_YEAR_STARTING_SUNDAY)


class Year(DateAccessor):
    @property
    def without_century(self) -> str:
        return self._date.strftime(FormatCode.YEAR.WITHOUT_CENTURY)

    @property
    def with_century(self) -> str:
        return self._date.strftime(FormatCode.YEAR.WITH_CENTURY)


class Month(DateAccessor):
    @property
    def short(self) -> str:
        return self._date.strftime(FormatCode.MONTH.SHORT)

    @property
    def full(self) -> str:
        return self._date.strftime(FormatCode.MONTH.FULL)

    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.MONTH.NUM)


class Day(DateAccessor):
    _suffixes = {day: suffix for days, suffix in [([1, 21, 31], "st"), ([2, 22], "nd"), ([3, 23], "rd")] for day in days}

    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.DAY.NUM)

    @property
    def of_year(self) -> str:
        return self._date.strftime(FormatCode.DAY.OF_YEAR)

    @property
    def suffix(self) -> str:
        return self._suffixes.get(self._date.day, 'th')

    @property
    def with_suffix(self) -> str:
        return f"{self._date.day}{self.suffix}"


class Hour(DateAccessor):
    @property
    def h24(self) -> str:
        return self._date.strftime(FormatCode.HOUR.H24)

    @property
    def h12(self) -> str:
        return self._date.strftime(FormatCode.HOUR.H12)

    @property
    def am_pm(self) -> str:
        return self._date.strftime(FormatCode.HOUR.AM_PM)


class Minute(DateAccessor):
    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.MINUTE.NUM)


class Second(DateAccessor):
    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.SECOND.NUM)


class MicroSecond(DateAccessor):
    @property
    def num(self) -> str:
        return self._date.strftime(FormatCode.MICROSECOND.NUM)



from __future__ import annotations

from collections.abc import Sequence
import json
from typing import Any, Iterable, Iterator, List, Callable

from django.utils.functional import cached_property

from maybe import Maybe

from .str import Accessor, Settings
from .translator import Translator


class SliceAccessor(Accessor):
    """An accessor class for all slicing-related Str methods"""

    class Settings(Settings):
        def __init__(self, raise_if_absent: bool = False) -> None:
            self.raise_if_absent = raise_if_absent

    settings = Settings()

    def __init__(self, parent: List_ = None) -> None:
        self.parent, self.settings = parent, self.Settings(raise_if_absent=type(self).settings.raise_if_absent)

    def __call__(self, parent: List_ = None, raise_if_absent: bool = None) -> SliceAccessor:
        self.parent, self.settings.raise_if_absent = Maybe(parent).else_(self.parent), Maybe(raise_if_absent).else_(self.settings.raise_if_absent)
        return self

    def before(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ before the given value. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(value, multiple_matches_forbidden=True)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[0]])

    def before_first(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ before the first instance of the given value."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[0]])

    def before_last(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ before the last instance of the given value."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[-1]])

    def after(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ after the given value. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(value, multiple_matches_forbidden=True)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[0]+1:])

    def after_first(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ after the first instance of the given value."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[0]+1:])

    def after_last(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ after the last instance of the given value."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[-1]+1:])

    def from_(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ from the given value onwards, including itself. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(value, multiple_matches_forbidden=True)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[0]:])

    def from_first(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ from the first instance of the given value onwards (including itself)."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[0]:])

    def from_last(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ from the last instance of the given value onwards (including itself)."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[matches[-1]:])

    def until(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ until the given value, including itself. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(value, multiple_matches_forbidden=True)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[0]+1])

    def until_first(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ until the first instance of the given value (including itself)."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[0]+1])

    def until_last(self, value: Any) -> List_:
        """Return all elements (if any) in the List_ until the last instance of the given value (including itself)."""
        matches = self._slice_helper(value, multiple_matches_forbidden=False)
        return type(self.parent)() if not matches else type(self.parent)(self.parent[:matches[-1]+1])

    def _slice_helper(self, value: Any, multiple_matches_forbidden: bool) -> List[int]:
        matches = [index for index, val in enumerate(self.parent) if val == value]

        if multiple_matches_forbidden:
            if len(matches) > 1:
                raise ValueError(f"Too many matches, return value would be ambigous (Expected 1, got {len(matches)}).")

        if self.settings.raise_if_absent and not matches:
            raise ValueError(f"'{value}' could not be found in '{self.parent}'.")

        return matches


class AttributeAccessor(Accessor):
    def __init__(self, parent: List_) -> None:
        self.parent = parent

    def __getattr__(self, attr: str) -> List_:
        return type(self.parent)([getattr(item, attr) for item in self.parent])


class ListSettings(Settings):
    def __init__(self) -> None:
        self.slice, self.translator, self.recursive = SliceAccessor(), Translator.default, True


class BaseList(list):
    """
    An alternative implementation of collections.UserList that inherits directly from 'list'. All the 'list' class inplace methods return self and therefore allow chaining when called from this class.
    """

    def __add__(self, other: list) -> BaseList:
        return type(self)(super().__add__(other))

    def __radd__(self, other: list) -> BaseList:
        return type(self)(other.__add__(self))

    def __iadd__(self, other):
        super().__iadd__(other)
        return self

    def __mul__(self, n: int) -> BaseList:
        return type(self)(super().__mul__(n))

    def __rmul__(self, n: int) -> BaseList:
        return type(self)(super().__rmul__(n))

    def __imul__(self, n: int) -> BaseList:
        super().__imul__(n)
        return self

    def append(self, item: Any) -> BaseList:
        """Same as list.append(), but returns self and thus allows chaining."""
        super().append(item)
        return self

    def extend(self, item: Any) -> BaseList:
        """Same as list.extend(), but returns self and thus allows chaining."""
        super().extend(item)
        return self

    def insert(self, index: int, item: Any) -> BaseList:
        """Same as list.insert(), but returns self and thus allows chaining."""
        super().insert(index, item)
        return self

    def remove(self, item: Any) -> BaseList:
        """Same as list.remove(), but returns self and thus allows chaining."""
        super().remove(item)
        return self

    def reverse(self) -> BaseList:
        """Same as list.reverse(), but returns self and thus allows chaining."""
        super().reverse()
        return self

    def sort(self, *args: Any, **kwargs: Any) -> BaseList:
        """Same as list.sort(), but returns self and thus allows chaining."""
        super().sort(*args, **kwargs)
        return self

    def clear(self) -> BaseList:
        """Same as list.clear(), but returns self and thus allows chaining."""
        super().clear()
        return self

    def copy(self):
        return type(self)(self)


class List_(BaseList):
    """
    Subclass of the builtin 'list' class with additional useful methods. All the 'list' class inplace methods return self and therefore allow chaining when called from this class.
    Recursively traverses its members and converts any str, list and dict instances into into their subtypes equivalents.
    """
    settings = ListSettings()

    def __init__(self, iterable: Iterable = None) -> None:
        super().__init__(iterable) if iterable is not None else super().__init__()

        if self.settings.recursive:
            for index, val in enumerate(self):
                self[index] = self.settings.translator.translate(val)

    @cached_property
    def slice(self) -> SliceAccessor:
        return SliceAccessor(parent=self)

    @cached_property
    def attr(self) -> AttributeAccessor:
        return AttributeAccessor(self)

    def apply(self, func: Callable) -> List_:
        return type(self)(map(func, self))

    def one(self) -> Any:
        if len(self) == 1:
            return self[0]
        else:
            raise ValueError(f"Expected '{self}' to contain a single value, but actual length was: {len(self)}.")

    def one_or_none(self) -> Any:
        if not self:
            return None
        elif len(self) == 1:
            return self[0]
        else:
            raise ValueError(f"Expected '{self}' to contain a single value or be empty, but actual length was: {len(self)}.")

    def split_into_batches(self, num_batches: int) -> Iterator[List_]:
        """Split this container into 'num_batches' equally sized containers of the same type. If the length of this container is not perfectly divisible by 'num_batches', the final container will be longer than the rest."""
        if num_batches >= len(self):
            yield type(self)(self)
        else:
            batch_size, final_batch_size_extra = len(self) // num_batches, len(self) % num_batches

            if not final_batch_size_extra:
                for run in range(0, len(self), batch_size):
                    yield self[run:run + batch_size]
            else:
                for run in range(0, (final_batch_position := batch_size*(num_batches - 1)), batch_size):
                    yield self[run:run + batch_size]

                yield self[final_batch_position:]

    def split_into_batches_of_size(self, batch_size: int) -> Iterator[List_]:
        """Split this container into smaller containers of the same type of size 'batch_size'. If the length of this container is not perfectly divisible by 'batch_size', the final container will be shorter than the rest."""
        if batch_size >= len(self):
            yield type(self)(self)
        else:
            for run in range(0, len(self), batch_size):
                yield self[run:run + batch_size]

    def to_json(self, indent: int = 4, **kwargs: Any) -> str:
        return json.dumps(self, indent=indent, **kwargs)

    def flatten(self) -> List_:
        """Recursively traverses any Sequence objects within this List_ and unpacks them in order into a new flat List_."""
        return self._flatten_more(iterable=self, output=type(self)())

    def _flatten_more(self, iterable: Iterable, output: List_) -> List_:
        for item in iterable:
            if isinstance(item, Sequence) and not isinstance(item, (str, bytes)):
                self._flatten_more(iterable=item, output=output)
            else:
                output.append(item)

        return output

    @classmethod
    def from_json(cls, json_string: str, **kwargs: Any) -> List_:
        item = json.loads(json_string, **kwargs)
        if isinstance(item, list):
            return cls(item)
        else:
            raise TypeError(f"The following json string resolves to type '{type(item).__name__}', not type '{list.__name__}':\n\n{json_string}")


Translator.translations[list] = List_



from __future__ import annotations

from typing import Any
import json
import simplejson

from requests import Session
import requests.models
from requests.compat import quote, quote_plus

from .enums import Enum
from .translator import Translator


class Response(requests.models.Response):
    """Subclass of requests.Response with a modified Response.json() method."""

    def __init__(self, namespace: dict) -> None:
        self.__dict__ = namespace

    def json(self) -> Any:
        """Returns Str, List_, and Dict_ items rather than their builtin superclasses. If there is no data will return None rather than raising JSONDecodeError."""
        try:
            return Translator.default.translate(super().json())
        except (json.JSONDecodeError, simplejson.JSONDecodeError):
            return None


class Http(Session):
    """
    Subclass of requests.Session which takes a 'base_url' constructor argument and prepends it to all future requests.
    It returns Str, List_, and Dict_ instances when deserializing json from responses and can automatically quote all urls passed to its http methods.
    """

    class QuoteLevel(Enum):
        NONE, NORMAL, PLUS = "none", "normal", "plus"

    def __init__(self, base_url: str = "", quote_level: Http.QuoteLevel = QuoteLevel.NONE) -> None:
        super().__init__()
        self.base_url, self.quote_level = base_url.strip('/'), quote_level

    def __repr__(self) -> str:
        return f"{type(self).__name__}(base_url={repr(self.base_url)}, auth={repr(self.auth)}, headers={repr(self.headers)})"

    def request(self, method: str, url: str, *args: Any, **kwargs: Any) -> Any:
        return Response(super().request(method=method, url=self._quote_encode(f"""{f"{self.base_url}/{url.strip('/')}".strip("/")}/"""), *args, **kwargs).__dict__)

    def _quote_encode(self, url: str) -> str:
        return self.QuoteLevel(self.quote_level).map_to({
            self.QuoteLevel.NONE: lambda url_: url_,
            self.QuoteLevel.NORMAL: lambda url_: quote(url_),
            self.QuoteLevel.PLUS: lambda url_: quote_plus(url_),
        })()



from __future__ import annotations

from typing import Any, Dict, Type
from json import loads


class Translator:
    translations: Dict[Type, Type] = {}
    default: Translator = None

    def __init__(self, translations: dict = None) -> None:
        self.translations = translations if translations is not None else self.translations.copy()

    def __call__(self, item: Any, recursive: bool = False) -> Any:
        return (self.translate_recursively if recursive else self.translate)(item)

    def translate(self, item: Any) -> Any:
        constructor = self.translations.get(type(item))
        return item if constructor is None else constructor(item)

    def translate_recursively(self, item: Any) -> Any:
        translated = self.translate(item)

        if isinstance(translated, list):
            for index, val in enumerate(translated):
                translated[index] = self.translate_recursively(val)
        elif isinstance(translated, dict):
            for key, val in translated.items():
                translated[key] = self.translate_recursively(val)

        return translated

    def translate_json(self, json: str, **kwargs: Any) -> Any:
        return self.translate(loads(json, **kwargs))


Translator.default = Translator(translations=Translator.translations)



from __future__ import annotations

import contextlib
import itertools
import io
import functools
import os
from typing import Any, Collection, List, Dict, Union, Type, Iterable, TypeVar, Callable, Iterator, cast, TYPE_CHECKING
import pathlib

import tabulate
import pandas as pd
from pandas.io.sql import SQLTable, pandasSQL_builder
from pandas.core.indexes.base import Index
import numpy as np
from maybe import Maybe

from .str import Str
from .enums import Enum
from .datetime import DateTime

if TYPE_CHECKING:
    from pathmagic import File


pd.set_option("max_columns", None)

PathLike = Union[str, os.PathLike, pathlib.Path]
FuncSig = TypeVar("FuncSig", bound=Callable)


def _check_import_is_available(func: FuncSig) -> FuncSig:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except ModuleNotFoundError as ex:
            exception = ModuleNotFoundError(f"Use of method '{func.__name__}' requires module '{ex.name}'. Please ensure it is available by installing it with pip.")
            exception.name, exception.path = ex.name, ex.path
            raise exception

    return cast(FuncSig, wrapper)


class Frame(pd.DataFrame):
    columns: Union[Index, Iterable]

    class InferRange(Enum):
        TRIM_SURROUNDING, STRIP_NULLS, SMALLEST_VALID = "trim_surrounding", "strip_nulls", "smallest_valid"

    class ColumnCase(Enum):
        IGNORE, SNAKE, CAMEL, PASCAL = "ignore", "snake", "camel", "pascal"

    class PathType(Enum):
        PATHMAGIC, PATHLIB, STRING = "pathmagic", "pathlib", "string"

    DEFAULT_COLUMN_CASE = ColumnCase.SNAKE
    DEFAULT_INFER_RANGE = InferRange.SMALLEST_VALID
    DEFAULT_PATH_TYPE = PathType.PATHMAGIC
    DEFAULT_SHEET_NAME = "Sheet1"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        with self._using_parent_constructor():
            super().__init__(*args, **kwargs)
            self._clean_dtypes()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(rows={len(self.index)}, columns={list(self.columns)})"

    @property
    def _constructor(self) -> Type[pd.DataFrame]:
        return type(self) if self._using_own_constructor() else pd.DataFrame

    def is_nan(self, val: Any) -> bool:
        return pd.isna(val)

    def to_excel(self, filepath: PathLike, sheet_name: str = None, index: bool = False, **kwargs: Any) -> PathLike:
        """Write this Frame to an xlsx file. Returns that File."""
        with ExcelWriter(filepath) as writer:
            self._write_to_excel(writer=writer, sheet_name=sheet_name, index=index, **kwargs)
        return self._get_path_constructor()(filepath)

    def to_sql(self, engine: Any, name: str, if_exists: str = "fail", index: bool = True, index_label: str = "id", primary_key: str = "id", schema: str = None, dtype: dict = None, **kwargs: Any) -> None:
        """Override of the pandas.DataFrame.to_sql() method allowing a primary key identity field to be supplied when creating the sql table."""
        table = SQLTable(name, pandasSQL_builder(engine), frame=self, index=index, if_exists=if_exists, index_label=index_label, keys=primary_key, schema=schema, dtype=dtype, **kwargs)
        table.create()
        table.insert(None)

    @_check_import_is_available
    def to_table(self, table: str, schema: str = None, database: str = None, if_exists: str = "fail", primary_key: str = "id", sql: Any = None) -> Any:
        """Load this Frame into a SQL database table using the config defaults of the sqlhandler library. An sqlhandler.Sql object can be provided to override the connection defaults."""
        from sqlhandler import Sql
        sql = sql if sql is not None else Sql(database=database)
        return sql.frame_to_table(self, table=table, schema=schema, if_exists=Sql.IfExists(if_exists), primary_key=primary_key)

    def to_dataframe(self) -> pd.DataFrame:
        """Convert this Frame to a pandas.DataFrame."""
        return pd.DataFrame(self)

    def sanitize_colnames(self, casing: str = None) -> Frame:
        """Strip newlines from the column names and apply arbitrary casing."""
        df: Frame = self.copy()
        df.columns = [str(colname).strip().replace("\n", "") for colname in df.columns]

        casing = Maybe(casing).else_(self.DEFAULT_COLUMN_CASE)
        if casing is not None:
            if casing == self.ColumnCase.SNAKE:
                df.columns = [Str(colname).case.snake() for colname in df.columns]
            elif casing == self.ColumnCase.CAMEL:
                df.columns = [Str(colname).case.camel() for colname in df.columns]
            elif casing == self.ColumnCase.PASCAL:
                df.columns = [Str(colname).case.pascal() for colname in df.columns]
            else:
                self.ColumnCase(casing)

        return df

    def to_ascii(self, index: bool = False, fancy: bool = True) -> str:
        """Convert this Frame to an ascii representation."""
        return str(tabulate.tabulate(self, headers=self.columns, tablefmt="fancy_grid" if fancy else "grid", showindex="never" if not index else "default"))

    @_check_import_is_available
    def to_desktop_as_excel(self, name: str, with_timestamp: bool = True, index: bool = False, **kwargs: Any) -> File:
        """Save this Frame to the current user's desktop as an xlsx file. Returns that File."""
        from pathmagic import Dir

        file = Dir.from_desktop().new_file(f"{name}_{DateTime.today().to_filetag() if with_timestamp else ''}", "xlsx")
        self.to_excel(file.path, index=index, **kwargs)
        return file

    # noinspection PyMethodOverriding
    def pivot(self, field_col: Union[pd.Series, str], value_col: Union[pd.Series, str]) -> Frame:
        """Pivot the content of this Frame."""
        field_name = field_col.name if isinstance(field_col, pd.Series) else field_col
        value_name = value_col.name if isinstance(value_col, pd.Series) else value_col

        pivoted: Frame = self.set_index([*[colname for colname in self.columns if colname not in {field_name, value_name}], field_name]).unstack(level=field_name).reset_index()
        pivoted.columns = [list(itertools.takewhile(lambda val: val, item))[-1] for item in pivoted.columns.to_flat_index()]
        return pivoted

    def unpivot(self, index_cols: List[Union[pd.Series, str]] = None, cols_to_unpivot: List[Union[pd.Series, str]] = None, field_name: str = "field", value_name: str = "value") -> Frame:
        """Unpivot the content of this Frame."""
        indexed = self.set_index(index_cols) if index_cols is not None else self.copy()

        if cols_to_unpivot is not None:
            indexed.drop([colname for colname in indexed.columns if colname not in cols_to_unpivot], axis=1, inplace=True)

        unpivoted = type(self)(indexed.stack().reset_index())
        final: Frame = unpivoted.rename(columns={unpivoted.columns[-2]: field_name, unpivoted.columns[-1]: value_name})
        final[field_name] = final[field_name].replace(True, 1).replace(False, 0)
        final.dropna(how="all", subset=list(final.columns[:-2]), inplace=True)

        return final

    def infer_dtypes(self) -> Frame:
        """Return a new Frame with newly inferred dtypes."""
        return type(self)(self.to_dict())

    def fillna_as_none(self) -> Frame:
        """Fill any nan values with None. Returns self."""
        df = self.copy()
        for name, col in df.iteritems():
            if col.isnull().any():
                df[name] = col.astype(object)

        return cast(Frame, df.where(df.notnull(), None))

    @_check_import_is_available
    def profile_report(self, *args: Any, style: dict = None, **kwargs: Any) -> Any:
        """Produce and return a pandas profile report"""
        import pandas_profiling

        style = Maybe(style).else_({'full_width': True})
        return pandas_profiling.ProfileReport(pd.DataFrame(self), *args, style=style, **kwargs)

    def profile_report_to(self, path: PathLike, *args: Any, **kwargs: Any) -> PathLike:
        """Produce a pandas profile report and return it as an html file"""
        file = self._get_path_constructor()(path)
        self.profile_report(*args, **kwargs).to_file(output_file=str(file))

        return file

    @classmethod
    def many_to_excel(cls, frames: Union[Collection[Frame], Dict[str, Collection[Frame]]], filepath: os.PathLike, index: bool = False, **kwargs: Any) -> PathLike:
        """Write an iterable of Frames or a mapping of string-keys and Frame-values into an xlsx file with several sheets."""

        named_frames = frames if isinstance(frames, dict) else {f"Sheet{idx + 1}": frame for idx, frame in enumerate(frames)}

        with ExcelWriter(filepath=filepath) as writer:
            for name, frame in named_frames.items():
                frame._write_to_excel(writer=writer, sheet_name=name, index=index, **kwargs)

        return cls._get_path_constructor()(filepath)

    @classmethod
    def from_excel(cls, filepath: os.PathLike, casing: str = None, skipcols: int = 0, infer_headers: bool = True, infer_range: str = None, password: str = None, **kwargs: Any) -> Frame:
        """Reads in the specified Excel spreadsheet into a Frame. Passes on arguments to the pandas read_excel function. Optionally changes casing on column names and strips out non-ascii characters."""

        filepath = os.fspath(filepath)

        if password is None:
            frame = cls(pd.read_excel(filepath, **kwargs))
        else:
            with cls._decrypted_stream_from_xlsx(path=filepath, password=password) as stream:
                frame = cls(pd.read_excel(stream, **kwargs))

        if skipcols:
            frame = frame.iloc[:, skipcols:]

        if infer_headers:
            frame._infer_column_headers()

        frame = frame.sanitize_colnames(casing=Maybe(casing).else_(cls.DEFAULT_COLUMN_CASE))

        infer_range = Maybe(infer_range).else_(cls.DEFAULT_INFER_RANGE)
        if infer_range:
            frame._infer_range(mode=infer_range)

        frame._infer_boolean_columns()
        return frame

    @classmethod
    def from_csv(cls, filepath: os.PathLike, casing: str = None, skipcols: int = 0, **kwargs: Any) -> Frame:
        """Reads in the specified csv file into a Frame. Passes on arguments to the pandas read_csv function. Optionally changes casing on column names and strips out non-ascii characters."""

        frame = cls(pd.read_csv(os.fspath(filepath), **kwargs))

        if skipcols:
            frame = frame.iloc[:, skipcols:]

        frame._infer_boolean_columns()
        return frame.sanitize_colnames(casing=casing)

    @classmethod
    def from_object(cls, obj: Any, private: bool = False) -> Frame:
        """Create a Frame from a single python object"""
        mappings = vars(obj) if private else {name: attr for name, attr in vars(obj).items() if not name.startswith("_")}
        return cls({name: [val] for name, val in mappings.items()})

    @classmethod
    def from_objects(cls, objects: Iterable[Any], private: bool = False) -> Frame:
        """Create a Frame from a homogenous list of python objects"""
        attrs = list(dict.fromkeys([name for obj in objects for name in vars(obj)]))
        valid_attrs = attrs if private else [name for name in attrs if not name.startswith("_")]
        return cls([tuple(vars(obj).get(attr) for attr in valid_attrs) for obj in objects], columns=valid_attrs)

    def _using_own_constructor(self) -> bool:
        return bool(object.__getattribute__(self, "_own_constructor_"))

    def _use_own_constructor(self, own: bool) -> None:
        object.__setattr__(self, "_own_constructor_", own)

    @contextlib.contextmanager
    def _using_parent_constructor(self) -> Iterator[None]:
        self._use_own_constructor(False)
        yield
        self._use_own_constructor(True)

    @classmethod
    @_check_import_is_available
    def _get_path_constructor(cls) -> Callable[..., PathLike]:
        if cls.DEFAULT_PATH_TYPE == Frame.PathType.PATHMAGIC:
            from pathmagic import File
            return File.from_pathlike
        elif cls.DEFAULT_PATH_TYPE == Frame.PathType.PATHLIB:
            return pathlib.Path
        elif cls.DEFAULT_PATH_TYPE == Frame.PathType.STRING:
            return os.fspath
        else:
            raise cls.PathType(cls.DEFAULT_PATH_TYPE)

    def _infer_column_headers(self) -> None:
        col_run = len(max("".join(["0" if self._value_is_null(col) else "1" for col in self.columns]).split("0")))

        longest_runs: Dict[int, Any] = {}
        for index, row in self.iterrows():
            longest_runs.setdefault(len(max("".join(["0" if self._value_is_null(col) else "1" for col in row]).split("0"))), []).append(index)

        if max(longest_runs) > col_run:
            first_longest_index = longest_runs[max(longest_runs)][0]
            self.columns = [str(val) for val in self.loc[first_longest_index]]
            self.drop(self.loc[:first_longest_index].index.tolist(), axis=0, inplace=True)

    def _infer_range(self, mode: Frame.InferRange = None) -> None:
        if mode is None:
            pass

        self.InferRange(mode).map_to({
            self.InferRange.TRIM_SURROUNDING: self._trim_nulls_around_table,
            self.InferRange.STRIP_NULLS: self._strip_fully_null,
            self.InferRange.SMALLEST_VALID: self._truncate_after_valid,
        })()

    def _trim_nulls_around_table(self) -> None:
        self._drop_rows_around_table()
        self._drop_columns_around_table()

    def _strip_fully_null(self) -> None:
        self._drop_fully_null_rows()
        self._drop_fully_null_columns()

    def _truncate_after_valid(self) -> None:
        self._drop_rows_around_table()
        self._truncate_rows_after_valid()

        self._drop_columns_around_table()
        self._truncate_columns_after_valid()

    def _drop_rows_around_table(self) -> None:
        nulls = [(index, row.isnull().all()) for index, row in self.iterrows()]
        invalid_rows: list = sum([list(itertools.takewhile(lambda row: row[1], iterable)) for iterable in (nulls, reversed(nulls))], [])
        if invalid_rows:
            self.drop([index for index, isnull in invalid_rows], axis=0, inplace=True)

    def _drop_columns_around_table(self) -> None:
        nulls = [(name, col.isnull().all()) for name, col in self.iteritems()]
        invalid_cols: list = sum([list(itertools.takewhile(lambda col: self._value_is_null(col[0]) and col[1], iterable)) for iterable in (nulls, reversed(nulls))], [])
        if invalid_cols:
            self.drop([name for name, isnull in invalid_cols], axis=1, inplace=True)

    def _truncate_rows_after_valid(self) -> None:
        empty_rows = self[self.isnull().all(axis=1) == True].index.tolist()
        if empty_rows:
            self.drop(self.loc[empty_rows[0]:].index.tolist(), axis=0, inplace=True)

    def _truncate_columns_after_valid(self) -> None:
        invalid_columns = list(itertools.dropwhile(lambda val: not self._value_is_null(val), self.columns))
        if invalid_columns:
            self.drop(invalid_columns, axis=1, inplace=True)

    def _drop_fully_null_rows(self) -> None:
        self.dropna(axis=0, how="all", inplace=True)

    def _drop_fully_null_columns(self) -> None:
        self.drop([name for name, col in self.iteritems() if self._value_is_null(name) and col.isnull().all()], axis=1, inplace=True)

    def _clean_dtypes(self) -> Frame:
        self._clean_nullable_ints()
        return self

    def _clean_nullable_ints(self) -> None:
        for name, col in self.iteritems():
            if col.dtype.name == "float64":
                if col.apply(lambda val: val is None or np.isnan(val) or val.is_integer()).all():
                    self[name] = col.astype("Int64")

    def _infer_boolean_columns(self) -> None:
        for name, col in self.iteritems():
            try:
                if col.isin([True, False, None]).all():
                    unfinished = col.replace(1, True).replace(0, False)
                    self[name] = unfinished.where(unfinished.notnull(), None)
            except (TypeError, SystemError):
                pass

    def _write_to_excel(self, writer: ExcelWriter, sheet_name: str, index: bool, **kwargs: Any) -> None:
        df, sheet_name = self.infer_dtypes(), Maybe(sheet_name).else_(self.DEFAULT_SHEET_NAME)
        super(type(df), df).to_excel(writer.writer, sheet_name=sheet_name, index=index, **kwargs)
        df._table_from_sheet(writer[sheet_name])

    def _table_from_sheet(self, sheet: Any) -> None:
        for idx, col in enumerate(self):
            series = self[col]
            datalen, headerlen = series.astype(str).map(len).max(), len(str(series.name))
            max_len = datalen + 1 if datalen > headerlen + 3 else headerlen + 4
            sheet.set_column(idx, idx, max_len)
        sheet.add_table(0, 0, len(sheet.table) - 1, len(sheet.table[0]) - 1, {"columns": [{"header": str(name)} for name in self.columns], "style": "Table Style Light 9"})

    @staticmethod
    def _value_is_null(value: Any) -> bool:
        return any([str(value).lower().startswith("unnamed"), *[str(value).lower() == nullproxy for nullproxy in ("none", "nan", "null", "")]])

    @staticmethod
    def _decrypted_stream_from_xlsx(path: PathLike, password: str) -> io.BytesIO:
        import msoffcrypto

        with open(path, "rb") as source_stream:
            file = msoffcrypto.OfficeFile(source_stream)
            file.load_key(password=password)
            destination_stream = io.BytesIO()
            file.decrypt(destination_stream)

        return destination_stream


class ExcelWriter:
    def __init__(self, filepath: PathLike) -> None:
        self.writer = pd.ExcelWriter(os.fspath(filepath), engine="xlsxwriter", mode="w")

    def __enter__(self) -> ExcelWriter:
        return self

    def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
        self.writer.save()

    def __getitem__(self, sheet_name: str) -> Any:
        # if self.writer.sheets is None:
        # self.writer.add_worksheet(sheet_name)
        return self.writer.sheets.get(sheet_name)



from __future__ import annotations

from typing import Any

from bs4 import BeautifulSoup, Tag
from prettierfier import prettify_html

from .str import Str


class Html(BeautifulSoup):
    """A subclass of bs4.BeautifulSoup which uses 'html.parser' as its default parser."""

    _stack = []

    def __init__(self, html: str = None, features: str = "html.parser", **kwargs: Any) -> None:
        element_classes = kwargs.get("element_classes", {})
        element_classes[Tag] = Html.Tag
        kwargs["element_classes"] = element_classes

        super().__init__(markup=html or "", features=features, **kwargs)

    def __repr__(self, encoding="unicode-escape") -> str:
        return str(self)

    def __str__(self) -> str:
        return Str(prettify_html(self.prettify())).re(multiline=True).sub(r"^( +)", lambda m: m.group(1)*4)

    def __enter__(self) -> Html:
        Html._stack.append(self)
        return self

    def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
        Html._stack.pop()

    def tag(self, name: str, content: str = None, / , attrs: dict = {}, **kwattrs: Any) -> Html.Tag:
        tag = self.new_tag(name=name, attrs=attrs, **kwattrs)

        if content:
            tag.string = content

        if Html._stack:
            Html._stack[-1].append(tag)
        else:
            self.append(tag)

        return tag

    class Tag(Tag):
        def __enter__(self) -> Html:
            Html._stack.append(self)
            return self

        def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
            Html._stack.pop()



from __future__ import annotations

import colour
from .enums import Enum


class ColorName(Enum):
    ALICE_BLUE = "aliceblue"
    ANTIQUE_WHITE = "antiquewhite"
    AQUA = "aqua"
    AQUAMARINE = "aquamarine"
    AZURE = "azure"
    BEIGE = "beige"
    BISQUE = "bisque"
    BLACK = "black"
    BLANCHED_ALMOND = "blanchedalmond"
    BLUE = "blue"
    BLUE_VIOLET = "blueviolet"
    BROWN = "brown"
    BURLYWOOD = "burlywood"
    CADET_BLUE = "cadetblue"
    CHARTREUSE = "chartreuse"
    CHOCOLATE = "chocolate"
    CORAL = "coral"
    CORNFLOWER_BLUE = "cornflowerblue"
    CORNSILK = "cornsilk"
    CRIMSON = "crimson"
    CYAN = "cyan"
    DARK_BLUE = "darkblue"
    DARK_CYAN = "darkcyan"
    DARK_GOLDENROD = "darkgoldenrod"
    DARK_GRAY = "darkgray"
    DARK_GREEN = "darkgreen"
    DARK_GREY = "darkgrey"
    DARK_KHAKI = "darkkhaki"
    DARK_MAGENTA = "darkmagenta"
    DARK_OLIVE_GREEN = "darkolivegreen"
    DARK_ORANGE = "darkorange"
    DARK_ORCHID = "darkorchid"
    DARK_RED = "darkred"
    DARK_SALMON = "darksalmon"
    DARK_SEA_GREEN = "darkseagreen"
    DARK_SLATE_BLUE = "darkslateblue"
    DARK_SLATE_GRAY = "darkslategray"
    DARK_SLATE_GREY = "darkslategrey"
    DARK_TURQUOISE = "darkturquoise"
    DARK_VIOLET = "darkviolet"
    DEEP_PINK = "deeppink"
    DEEP_SKY_BLUE = "deepskyblue"
    DIM_GRAY = "dimgray"
    DIM_GREY = "dimgrey"
    DODGER_BLUE = "dodgerblue"
    FIREBRICK = "firebrick"
    FLORAL_WHITE = "floralwhite"
    FOREST_GREEN = "forestgreen"
    FUCHSIA = "fuchsia"
    GAINSBORO = "gainsboro"
    GHOST_WHITE = "ghostwhite"
    GOLD = "gold"
    GOLDENROD = "goldenrod"
    GRAY = "gray"
    GREEN = "green"
    GREEN_YELLOW = "greenyellow"
    GREY = "grey"
    HONEYDEW = "honeydew"
    HOT_PINK = "hotpink"
    INDIAN_RED = "indianred"
    INDIGO = "indigo"
    IVORY = "ivory"
    KHAKI = "khaki"
    LAVENDER = "lavender"
    LAVENDER_BLUSH = "lavenderblush"
    LAWN_GREEN = "lawngreen"
    LEMON_CHIFFON = "lemonchiffon"
    LIGHT_BLUE = "lightblue"
    LIGHT_CORAL = "lightcoral"
    LIGHT_CYAN = "lightcyan"
    LIGHT_GOLDENROD = "lightgoldenrod"
    LIGHT_GOLDENROD_YELLOW = "lightgoldenrodyellow"
    LIGHT_GRAY = "lightgray"
    LIGHT_GREEN = "lightgreen"
    LIGHT_GREY = "lightgrey"
    LIGHT_PINK = "lightpink"
    LIGHT_SALMON = "lightsalmon"
    LIGHT_SEA_GREEN = "lightseagreen"
    LIGHT_SKY_BLUE = "lightskyblue"
    LIGHT_SLATE_BLUE = "lightslateblue"
    LIGHT_SLATE_GRAY = "lightslategray"
    LIGHT_SLATE_GREY = "lightslategrey"
    LIGHT_STEEL_BLUE = "lightsteelblue"
    LIGHT_YELLOW = "lightyellow"
    LIME = "lime"
    LIME_GREEN = "limegreen"
    LINEN = "linen"
    MAGENTA = "magenta"
    MAROON = "maroon"
    MEDIUM_AQUAMARINE = "mediumaquamarine"
    MEDIUM_BLUE = "mediumblue"
    MEDIUM_ORCHID = "mediumorchid"
    MEDIUM_PURPLE = "mediumpurple"
    MEDIUM_SEA_GREEN = "mediumseagreen"
    MEDIUM_SLATE_BLUE = "mediumslateblue"
    MEDIUM_SPRING_GREEN = "mediumspringgreen"
    MEDIUM_TURQUOISE = "mediumturquoise"
    MEDIUM_VIOLET_RED = "mediumvioletred"
    MIDNIGHT_BLUE = "midnightblue"
    MINT_CREAM = "mintcream"
    MISTY_ROSE = "mistyrose"
    MOCCASIN = "moccasin"
    NAVAJO_WHITE = "navajowhite"
    NAVY = "navy"
    NAVY_BLUE = "navyblue"
    OLD_LACE = "oldlace"
    OLIVE = "olive"
    OLIVE_DRAB = "olivedrab"
    ORANGE = "orange"
    ORANGE_RED = "orangered"
    ORCHID = "orchid"
    PALE_GOLDENROD = "palegoldenrod"
    PALE_GREEN = "palegreen"
    PALE_TURQUOISE = "paleturquoise"
    PALE_VIOLET_RED = "palevioletred"
    PAPAYA_WHIP = "papayawhip"
    PEACH_PUFF = "peachpuff"
    PERU = "peru"
    PINK = "pink"
    PLUM = "plum"
    POWDER_BLUE = "powderblue"
    PURPLE = "purple"
    RED = "red"
    ROSY_BROWN = "rosybrown"
    ROYAL_BLUE = "royalblue"
    SADDLE_BROWN = "saddlebrown"
    SALMON = "salmon"
    SANDY_BROWN = "sandybrown"
    SEASHELL = "seashell"
    SEA_GREEN = "seagreen"
    SIENNA = "sienna"
    SILVER = "silver"
    SKY_BLUE = "skyblue"
    SLATE_BLUE = "slateblue"
    SLATE_GRAY = "slategray"
    SLATE_GREY = "slategrey"
    SNOW = "snow"
    SPRING_GREEN = "springgreen"
    STEEL_BLUE = "steelblue"
    TAN = "tan"
    THISTLE = "thistle"
    TOMATO = "tomato"
    TURQUOISE = "turquoise"
    VIOLET = "violet"
    VIOLET_RED = "violetred"
    WHEAT = "wheat"
    WHITE = "white"
    WHITE_SMOKE = "whitesmoke"
    YELLOW = "yellow"
    YELLOW_GREEN = "yellowgreen"


class Color(colour.Color):
    """Subclass of colour.Color with access to a Color.Name emum of color names that are valid to be passed into the Color constructor."""
    Name = ColorName



from __future__ import annotations

from typing import List, Any, Type
from collections.abc import Mapping
import json

from django.utils.functional import cached_property

from .str import Str, Accessor
from .str import RegexAccessor as StrRegexAccessor, Settings
from .translator import Translator

from maybe import Maybe


def is_special_private(name: str) -> bool:
    return name.startswith("_") and name.endswith("_")


def is_valid_for_attribute_actions(name: Any, dict_class: Type[Dict_]) -> bool:
    return isinstance(name, str) and name not in dict_class.settings.dict_fields and not is_special_private(name) and name.isidentifier()


class AccessError(KeyError, AttributeError):
    pass


class RegexAccessor(Accessor):
    """An accessor class for all regex-related Dict_ methods"""
    settings = StrRegexAccessor.Settings()

    def __init__(self, parent: Dict_ = None) -> None:
        default = type(self).settings
        self.parent, self.settings = parent, StrRegexAccessor.Settings(dotall=default.dotall, ignorecase=default.ignorecase, multiline=default.multiline)
        self.str = Str()

    def __call__(self, parent: Dict_ = None, dotall: bool = None, ignorecase: bool = None, multiline: bool = None) -> RegexAccessor:
        self.parent = Maybe(parent).else_(self.parent)
        self.settings.dotall = Maybe(dotall).else_(self.settings.dotall)
        self.settings.ignorecase = Maybe(ignorecase).else_(self.settings.ignorecase)
        self.settings.multiline = Maybe(multiline).else_(self.settings.multiline)
        return self

    def filter(self, regex: str) -> Dict_:
        """Remove any key-value pairs where the key is not a string, or where it is a string but doesn't match the given regex."""
        return type(self.parent)(
            {key: val for key, val in self.parent.items()
             if isinstance(key, str) and Str(key).re(dotall=self.settings.dotall, ignorecase=self.settings.ignorecase, multiline=self.settings.multiline).search(regex) is not None}
        )

    def get_all(self, regex: str, limit: int = None) -> List[Any]:
        """Return a list of all the values whose keys match the given regex."""
        vals = self.filter(regex)

        if limit is not None and len(vals) > limit:
            raise KeyError(f"Got {len(vals)} matches: {', '.join([repr(val) for val in vals])}. Expected at most {limit} match(es).")
        else:
            return list(vals.values())

    def get_one(self, regex: str) -> Any:
        """Return the value whose key matches the given regex. KeyError will be raised if multiple matches are found."""
        return self.get_all(regex=regex, limit=1)[0]


class BaseDict(dict):
    """
    An alternative implementation of collections.UserDict that inherits directly from 'dict'. All the 'dict' class inplace methods return self and therefore allow chaining when called from this class.
    """

    def update(self, item: Mapping = None, **kwargs) -> BaseDict:
        """Same as dict.update(), but returns self and thus allows chaining."""
        super().update(item)
        return self

    def clear(self) -> BaseDict:
        """Same as dict.clear(), but returns self and thus allows chaining."""
        super().clear()
        return self

    def copy(self) -> BaseDict:
        return type(self)(self.copy())


class Dict_(BaseDict):
    """
    Subclass of the builtin 'dict' class with where inplace methods like dict.update() return self and therefore allow chaining.
    Also allows item access dynamically through attribute access. It recursively converts any str, list, and dict instances into Str, List_, and Dict_.
    """

    class Settings(Settings):
        def __init__(self) -> None:
            self.re, self.dict_fields, self.translator, self.recursive = RegexAccessor.settings, {attr for attr in [*dir(dict()), "settings"] if not attr.startswith("_")}, Translator.default, True

    settings = Settings()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if self.settings.recursive:
            for key, val in self.items():
                self[key] = val

    def __getitem__(self, name: str) -> Any:
        try:
            return super().__getitem__(name)
        except KeyError:
            self[name] = default = self._factory_(name=name)
            return default

    def __setitem__(self, name: str, val: Any) -> None:
        clean_val = self.settings.translator.translate(val) if self.settings.recursive else val

        super().__setitem__(name, clean_val)
        if is_valid_for_attribute_actions(name, type(self)):
            super().__setattr__(name, clean_val)

    def __delitem__(self, name: str) -> None:
        super().__delitem__(name)

        if is_valid_for_attribute_actions(name, type(self)):
            super().__delattr__(name)

    def __getattr__(self, name: str) -> Dict_:
        if is_special_private(name):
            raise AttributeError(name)
        else:
            return self[name]

    def __setattr__(self, name: str, val: Any) -> None:
        if is_special_private(name):
            super().__setattr__(name, val)
        else:
            self[name] = val

    def __delattr__(self, name: str) -> None:
        super().__delattr__(name)

        if not is_special_private(name):
            super().__delitem__(name)

    def _factory_(self, name: str) -> Dict_:
        raise AccessError(f"'{name}' not found in {type(self).__name__}: {self}")

    @cached_property
    def re(self) -> RegexAccessor:
        return RegexAccessor(parent=self)

    def to_json(self, indent: int = 4, **kwargs: Any) -> str:
        return json.dumps(self, indent=indent, **kwargs)

    @classmethod
    def from_json(cls, json_string: str, **kwargs: Any) -> Dict_:
        if isinstance((item := json.loads(json_string, **kwargs)), dict):
            return cls(item)
        else:
            raise TypeError(f"The following json string resolves to type '{type(item).__name__}', not type '{dict.__name__}':\n\n{json_string}")


class DefaultDict(Dict_):
    def _factory_(self, name: str) -> DefaultDict:
        return type(self)()


Translator.translations[dict] = Dict_



from __future__ import annotations

import copy
import itertools
import re
from typing import Any, Callable, Iterator, Iterable, List, Tuple, Dict, Mapping, Match, Union
import warnings

import regex as regexmod
import case_conversion
import inflect
import clipboard
from django.utils.functional import cached_property

from maybe import Maybe

from .enums import Enum
from .translator import Translator

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from fuzzywuzzy import fuzz


class Accessor:
    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"


class Settings:
    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    def deepcopy(self) -> Settings:
        return copy.deepcopy(self)


class RegexAccessor(Accessor):
    """An accessor class for all regex-related Str methods"""

    class Settings(Settings):
        def __init__(self, dotall: bool = True, ignorecase: bool = True, multiline: bool = False):
            self.dotall, self.ignorecase, self.multiline = dotall, ignorecase, multiline

        def __repr__(self) -> str:
            return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

        def __int__(self) -> int:
            return self.to_flag()

        def __and__(self, other: Union[int, re.RegexFlag]) -> int:
            return self.to_flag() & other

        def __or__(self, other: Union[int, re.RegexFlag]) -> int:
            return self.to_flag() | other

        def __rand__(self, other: Union[int, re.RegexFlag]) -> int:
            return self.to_flag() & other

        def __ror__(self, other: Union[int, re.RegexFlag]) -> int:
            return self.to_flag() | other

        def to_flag(self) -> re.RegexFlag:
            ret = re.RegexFlag(0)
            for attr, flag in (self.dotall, re.DOTALL), (self.ignorecase, re.IGNORECASE), (self.multiline, re.MULTILINE):
                if attr:
                    ret |= flag
            return ret

    settings = Settings()

    def __init__(self, parent: Str = None) -> None:
        self.parent, self.settings = parent, self.settings.deepcopy()

    def __call__(self, parent: Str = None, dotall: bool = None, ignorecase: bool = None, multiline: bool = None) -> RegexAccessor:
        self.parent = Maybe(parent).else_(self.parent)
        self.settings.dotall = Maybe(dotall).else_(self.settings.dotall)
        self.settings.ignorecase = Maybe(ignorecase).else_(self.settings.ignorecase)
        self.settings.multiline = Maybe(multiline).else_(self.settings.multiline)
        return self

    def search(self, regex: str, **kwargs: Any) -> Match[str]:
        """Perform a regex.search on this Str"""
        ret: Match[str] = regexmod.search(regex, self.parent, flags=Maybe(kwargs.pop("flags", None)).else_(self.settings.to_flag()), **kwargs)
        return ret

    def sub(self, regex: str, sub: Union[str, Callable], raise_for_failure: bool = False, **kwargs: Any) -> Str:
        """Perform a regex.search on this Str"""
        subbed = regexmod.sub(regex, sub, self.parent, flags=Maybe(kwargs.pop("flags", None)).else_(self.settings.to_flag()), **kwargs)

        if raise_for_failure:
            if self.parent == subbed:
                raise ValueError(f"'{self.parent}' was not changed!")

        return type(self.parent)(subbed)

    def finditer(self, regex: str, **kwargs: Any) -> Iterator[Match[str]]:
        """Perform a regex.finditer on this Str"""
        ret: Iterator[Match[str]] = regexmod.finditer(regex, self.parent, flags=Maybe(kwargs.pop("flags", None)).else_(self.settings.to_flag()), **kwargs)
        return ret

    def split(self, regex: str, **kwargs: Any) -> List[Str]:
        """Perform a regex.split on this Str"""
        return [type(self.parent)(item) for item in regexmod.split(regex, self.parent, flags=Maybe(kwargs.pop("flags", None)).else_(self.settings.to_flag()), **kwargs)]

    def escape(self) -> str:
        """Perform a re.escape on this Str"""
        return type(self.parent)(re.escape(self.parent))


class FuzzyAccessor(Accessor):
    """An accessor class for all fuzzy-matching-related Str methods"""

    class Settings(Settings):
        def __init__(self, tokenize: bool = False, partial: bool = False) -> None:
            self.tokenize, self.partial = tokenize, partial

    settings = Settings()

    def __init__(self, parent: Str = None) -> None:
        self.parent, self.settings = parent, self.settings.deepcopy()
        self._determine_matcher()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    def __call__(self, parent: Str = None, tokenize: bool = None, partial: bool = None) -> FuzzyAccessor:
        self.parent = Maybe(parent).else_(self.parent)
        self.settings.tokenize = Maybe(tokenize).else_(self.settings.tokenize)
        self.settings.partial = Maybe(partial).else_(self.settings.partial)
        self._determine_matcher()
        return self

    def match(self, other: str) -> int:
        """Return a score out of 100 representing a fuzzy-match between this Str and another using the current fuzzy-matching settings"""
        return self._matcher(self.parent, other)

    def best_n_matches(self, possible_matches: List[str], num: int = 3) -> Dict[str, int]:
        """Return a number of the best fuzzy matches between this Str and an iterable of strings in descending order using the current fuzzy-matching settings"""
        match_scores = {self.match(match): match for match in possible_matches}
        return {match_scores[score]: score for index, score in itertools.takewhile(lambda tup: tup[0] < num, enumerate(sorted(match_scores, reverse=True)))}

    def _determine_matcher(self) -> None:
        for func, tokenize, partial in [(fuzz.ratio, False, False), (fuzz.partial_ratio, False, True), (fuzz.token_set_ratio, True, False), (fuzz.partial_token_set_ratio, True, True)]:
            if self.settings.tokenize is tokenize and self.settings.partial is partial:
                self._matcher = func
                return


class CasingAccessor(Accessor):
    """An accessor class for all casing-related Str methods"""

    class Settings(Settings):
        def __init__(self, detect_acronyms: bool = True, acronyms: list = None) -> None:
            self.detect_acronyms, self.acronyms = detect_acronyms, acronyms

    settings = Settings()

    def __init__(self, parent: Str = None) -> None:
        self.parent, self.settings = parent, self.settings.deepcopy()

    def __call__(self, parent: Str = None, detect_acronyms: bool = None, acronyms: list = None) -> CasingAccessor:
        self.parent = Maybe(parent).else_(self.parent)
        self.settings.detect_acronyms = Maybe(detect_acronyms).else_(self.settings.detect)
        self.settings.acronyms = Maybe(acronyms).else_(self.settings.acronyms)
        return self

    def snake(self) -> Str:
        """snake_case this Str"""
        return type(self.parent)(case_conversion.snakecase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms).strip("_"))

    def camel(self) -> Str:
        """camelCase this Str"""
        return type(self.parent)(case_conversion.camelcase(self.snake(), detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def pascal(self) -> Str:
        """PascalCase this Str"""
        return type(self.parent)(case_conversion.pascalcase(self.snake(), detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def dash(self) -> Str:
        """dash-case this Str"""
        return type(self.parent)(case_conversion.dashcase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def constant(self) -> Str:
        """CONSTANT_CASE this Str"""
        return type(self.parent)(case_conversion.constcase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def dot(self) -> Str:
        """dot.case this Str"""
        return type(self.parent)(case_conversion.dotcase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def slash(self) -> Str:
        """slash/case this Str"""
        return type(self.parent)(case_conversion.slashcase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def backslash(self) -> Str:
        """backslash\\case this Str"""
        return type(self.parent)(case_conversion.backslashcase(self.parent, detect_acronyms=self.settings.detect_acronyms, acronyms=self.settings.acronyms))

    def identifier(self) -> Str:
        """Turn this Str into a valid python identifier by first snake_casing it and then stripping away invalid characters"""
        return self.snake().re.sub(r"^(?=\d+)", "_")

    def plural(self) -> Str:
        """Return the English plural of this Str"""
        return type(self.parent)(inflect.engine().plural(self.parent))

    def from_enum(self, case: Str.Case) -> Str:
        return getattr(self, str(case))()


class SliceAccessor(Accessor):
    """An accessor class for all slicing-related Str methods"""

    class Settings(Settings):
        def __init__(self, raise_if_absent: bool = False) -> None:
            self.raise_if_absent = raise_if_absent

    settings = Settings()

    def __init__(self, parent: Str = None) -> None:
        self.parent, self.settings = parent, self.settings.deepcopy()

    def __call__(self, parent: Str = None, raise_if_absent: bool = None) -> SliceAccessor:
        self.parent, self.settings.raise_if_absent = Maybe(parent).else_(self.parent), Maybe(raise_if_absent).else_(self.settings.raise_if_absent)
        return self

    def before(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str before the given regex. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=True)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[0].span()[0]])

    def before_first(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str before the first instance of the given regex."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[0].span()[0]])

    def before_last(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str before the last instance of the given regex."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[-1].span()[0]])

    def after(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str after the given regex. Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=True)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[0].span()[1]:])

    def after_first(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str after the first instance of the given regex."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[0].span()[1]:])

    def after_last(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str after the last instance of the given regex."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[-1].span()[1]:])

    def from_(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str from the given regex onwards (including itself). Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=True)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[0].span()[0]:])

    def from_first(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str from the first instance of the given regex onwards (including itself)."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[0].span()[0]:])

    def from_last(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str from the last instance of the given regex onwards (including itself)."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[matches[-1].span()[0]:])

    def until(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str until given regex (including itself). Raises ValueError if multiple matches are found."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=True)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[0].span()[1]])

    def until_first(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str until the first instance of the given regex (including itself)."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[0].span()[1]])

    def until_last(self, regex: str) -> Str:
        """Return a new Str from the portion of this Str until the last instance of the given regex (including itself)."""
        matches = self._slice_helper(regex, multiple_matches_forbidden=False)
        return type(self.parent)("") if not matches else type(self.parent)(self.parent[:matches[-1].span()[1]])

    def _slice_helper(self, regex: str, multiple_matches_forbidden: bool) -> List[Match[str]]:
        matches = list(self.parent.re.finditer(regex=regex))

        if multiple_matches_forbidden:
            if len(matches) > 1:
                raise ValueError(f"Too many matches for regex pattern '{regex}' in '{self.parent}', return value would be ambigous (Expected 1, got {len(matches)}).")

        if self.settings.raise_if_absent and not matches:
            raise ValueError(f"'{regex}' could not be found in '{self.parent}'.")

        return matches


class TrimAccessor(Accessor):
    """An accessor class for all stripping-related Str methods"""

    def __init__(self, parent: Str = None) -> None:
        self.parent = parent

    def __call__(self, parent: Str = None) -> TrimAccessor:
        self.parent = Maybe(parent).else_(self.parent)
        return self

    def all_whitespace(self) -> Str:
        """Strip away all whitespace"""
        return self.parent.re.sub(r"\s+", "").strip()

    def whitespace_runs(self, newlines: int = 1, tabs: int = 1, spaces: int = 1) -> Str:
        """Replace all whitespace runs with a single instance of the appropriate whitespace character space"""
        return self.parent.re.sub(r"[ \t]*\n[ \t]*", "\n").re.sub(r" *\t *", "\t").re.sub(fr"\n{{{newlines+1},}}", "\n"*newlines).re.sub(fr"\t{{{tabs+1},}}", "\t"*tabs).re.sub(fr" {{{spaces+1},}}", " "*spaces).strip()

    def non_alphanumeric(self) -> Str:
        """Strip away all non-alphanumeric characters"""
        return self.parent.re.sub(r"[^A-Za-z0-9]", "")

    def non_ascii(self) -> Str:
        """Strip away all non-ascii characters"""
        return type(self.parent)(self.parent.encode("ascii", errors="ignore").decode("UTF-8"))


class StrSettings(Settings):
    def __init__(self) -> None:
        self.re, self.case, self.slice, self.fuzzy = RegexAccessor.settings, CasingAccessor.settings, SliceAccessor.settings, FuzzyAccessor.settings


class BaseStr(str):
    """An alternative implementation of collections.UserString that inherits directly from 'str'."""

    def __add__(self, other: str) -> BaseStr:
        return type(self)(super().__add__(other))

    def __radd__(self, other: str) -> BaseStr:
        return type(self)(other) + self

    def __mul__(self, n) -> BaseStr:
        return type(self)(super().__mul__(n))

    def __rmul__(self, n) -> BaseStr:
        return self*n

    def __mod__(self, args: Any) -> BaseStr:
        return type(self)(super().__mod__(args))

    def __rmod__(self, template: str) -> BaseStr:
        return type(self)(template) % self

    def capitalize(self) -> BaseStr:
        return type(self)(super().capitalize())

    def casefold(self) -> BaseStr:
        return type(self)(super().casefold())

    def center(self, width, *args) -> BaseStr:
        return type(self)(super().center(width, *args))

    def expandtabs(self, tabsize: int = 8) -> BaseStr:
        return type(self)(super().expandtabs(tabsize))

    def format(self, *args: Any, **kwds: Any) -> BaseStr:
        return type(self)(super().format(*args, **kwds))

    def format_map(self, mapping: Mapping[str, Any]) -> BaseStr:
        return type(self)(super().format_map(mapping))

    def join(self, iterable: Iterable[str]) -> BaseStr:
        return type(self)(super().join(iterable))

    def ljust(self, width: int, fillchar: str = " ") -> BaseStr:
        return type(self)(super().ljust(width, fillchar))

    def lower(self) -> BaseStr:
        return type(self)(super().lower())

    def lstrip(self, chars: str = None) -> BaseStr:
        return type(self)(super().lstrip(chars))

    def partition(self, sep: str) -> Tuple[BaseStr, ...]:
        return tuple(type(self)(item) for item in super().partition(sep))

    def replace(self, old: str, new: str, maxsplit: int = -1) -> BaseStr:
        return type(self)(super().replace(old, new, maxsplit))

    def rjust(self, width: int, fillchar: str = " ") -> BaseStr:
        return type(self)(super().rjust(width, fillchar))

    def rpartition(self, sep: str) -> Tuple[BaseStr, ...]:
        return tuple(type(self)(item) for item in super().rpartition(sep))

    def rstrip(self, chars: str = None) -> BaseStr:
        return type(self)(super().rstrip(chars))

    def strip(self, chars: str = None) -> BaseStr:
        return type(self)(super().strip(chars))

    def swapcase(self) -> BaseStr:
        return type(self)(super().swapcase())

    def title(self) -> BaseStr:
        return type(self)(super().title())

    def translate(self, *args) -> BaseStr:
        return type(self)(super().translate(*args))

    def upper(self) -> BaseStr:
        return type(self)(super().upper())

    def zfill(self, width: int) -> BaseStr:
        return type(self)(super().zfill(width))


class Str(BaseStr):
    """A subclass of the builin 'str' class which supports inplace mutation using item access. Has additional methods and accessor objects with additional methods for casing, regex, fuzzy-matching, trimming, and slicing."""

    class Case(Enum):
        SNAKE, CAMEL, PASCAL, CONSTANT = CasingAccessor.snake.__name__, CasingAccessor.camel.__name__, CasingAccessor.pascal.__name__, CasingAccessor.constant.__name__
        DOT, DASH, SLASH, BACKSLASH = CasingAccessor.dot.__name__, CasingAccessor.dash.__name__, CasingAccessor.slash.__name__, CasingAccessor.backslash.__name__
        IDENTIFIER, PLURAL = CasingAccessor.identifier.__name__, CasingAccessor.plural.__name__

    settings = StrSettings()

    @cached_property
    def re(self) -> RegexAccessor:
        return RegexAccessor(parent=self)

    @cached_property
    def case(self) -> CasingAccessor:
        return CasingAccessor(parent=self)

    @cached_property
    def slice(self) -> SliceAccessor:
        return SliceAccessor(parent=self)

    @cached_property
    def trim(self) -> TrimAccessor:
        return TrimAccessor(parent=self)

    @cached_property
    def fuzzy(self) -> FuzzyAccessor:
        return FuzzyAccessor(parent=self)

    def to_clipboard(self) -> None:
        """Save the content of this string to the clipboard"""
        clipboard.copy(self)

    # parsing

    def find_all(self, substring: str, regex: bool = False, overlapping: bool = True, not_within: List[Tuple[str, str]] = None) -> List[int]:
        """Return a list of all the index positions at which the given substring occurs."""
        pattern = substring if regex else re.escape(substring)
        prefix = "".join([fr"{start}.*?{end}|" for start, end in not_within]) if not_within is not None else ""
        return [match.start() for match in regexmod.finditer(fr"{prefix}({pattern})", self, overlapped=overlapping if not_within is None else False, flags=self.re.settings.to_flag() if regex else 0) if match.group(1)]

    def extract_uk_postcode(self) -> Str:
        """Attempt to extract a valid UK postcode from the given string. Returns None upon failure."""
        match = self.re.search(r"([A-Za-z]{1,2}[0-9]{1,2}[A-Za-z]?)( )?([0-9][A-Za-z]{2})")
        return None if match is None else type(self)(f"{match.group(1).upper()} {match.group(3).upper()}")

    @classmethod
    def from_clipboard(cls) -> Str:
        """Create a Str from the content of the clipboard"""
        return cls(clipboard.paste())


Translator.translations[str] = Str



from __future__ import annotations

from typing import Any, List
import enum

import aenum


class EnumMeta(aenum.EnumMeta):
    # Implementation of EnumMeta.__new__() and EnumMeta.__prepare__() alongside subtypes.Enum inheriting from enum.Enum unnecessarily
    # when it already inherits from aenum.Enum which is a subclass of enum.Enum, is done so that the goddamn PyCharm type checker
    # is satisfied that subtypes.Enum is, in fact, a freaking enumeration. Otherwise it just doesn't get it.

    def __new__(mcs, name: str, bases: tuple, namespace: dict, init: Any = None, start: Any = None, settings: tuple = ()) -> EnumMeta:
        if enum.Enum in bases:
            (bases := list(bases)).remove(enum.Enum)
            bases = tuple(bases)

        return super().__new__(mcs, name, bases, namespace, init, start, settings)

    @classmethod
    def __prepare__(mcs, name: str, bases: tuple, init: Any = None, start: Any = None, settings: tuple = ()) -> EnumMeta:
        if enum.Enum in bases:
            (bases := list(bases)).remove(enum.Enum)
            bases = tuple(bases)

        return super().__prepare__(name, bases, init, start, settings)

    def __repr__(cls) -> str:
        return f"{cls.__name__}[{', '.join([f'{member.name}={repr(member.value)}' for member in cls])}]"

    def __str__(cls) -> str:
        return cls.__name__

    def __call__(cls, *args: Any, **kwargs: Any):
        try:
            return super().__call__(*args, **kwargs)
        except ValueError as ex:
            msg, = ex.args
            raise ValueError(f"{msg}, must be one of: {', '.join([repr(member.value) for member in cls])}.")

    @property
    def names(cls) -> List[str]:
        """A list of the names in this Enum."""
        return [member.name for member in cls]

    @property
    def values(cls) -> List[Any]:
        """A list of the values in this Enum."""
        return [member.value for member in cls]

    def extend(cls, name: str, value: Any) -> None:
        """Extend this Enum with an additional member created from the given name and value."""
        aenum.extend_enum(cls, name, value)

    def is_enum(cls, candidate: Any) -> bool:
        """Returns True if the candidate is a subclass of Enum, otherwise returns False."""
        try:
            return issubclass(candidate, enum.Enum)
        except TypeError:
            return False


class ValueEnumMeta(EnumMeta):
    def __getattribute__(cls, name: str) -> Any:
        value = super().__getattribute__(name)

        if isinstance(value, cls):
            value = value.value

        return value


class BaseEnum(aenum.Enum):
    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={self.name}, value={repr(self.value)})"

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: Any) -> bool:
        return other is self or other == self.value

    def __ne__(self, other: Any) -> bool:
        return other is not self and other != self.value

    def __str__(self) -> str:
        return str(self.value)

    def map_to(self, mapping: dict, raise_for_failure: bool = True) -> Any:
        if (ret := mapping.get(self)) is None and raise_for_failure:
            raise ValueError(f"No mapping for '{self}' found in {mapping}.")

        return ret


class Enum(BaseEnum, enum.Enum, metaclass=EnumMeta):
    """A subclass of aenum.Enum with additional methods."""


class ValueEnum(BaseEnum, metaclass=ValueEnumMeta):
    """A subclass of subtypes.Enum. Attribute access on descendants of this class returns the value corresponding to that name, rather than returning the member."""


class BaseAutoEnum(aenum.Enum):
    _settings_ = aenum.AutoValue

    def _generate_next_value_(name: str, start: str, count: str, last_values: List[str]) -> str:
        return name.lower()


class AutoEnum(Enum, BaseAutoEnum):
    """A subclass of subtypes.Enum. Automatically uses _generate_next_value_ when values are missing"""


class ValueAutoEnum(BaseEnum, BaseAutoEnum, metaclass=ValueEnumMeta):
    """A subclass of subtypes.ValueEnum. Missing values are automatically supplied by _generate_next_value_."""



from __future__ import annotations


class SingletonMeta(type):
    def __call__(cls) -> Singleton:
        if cls._instance_ is None:
            cls._instance_ = super().__call__()

        return cls._instance_


class Singleton(metaclass=SingletonMeta):
    """A singleton class that will always return the same object from its constuctor. Intended to be subclassed."""
    _instance_ = None

    def __repr__(self) -> str:
        return type(self).__name__



__all__ = [
    "Enum", "ValueEnum", "AutoEnum", "ValueAutoEnum",
    "Html",
    "Http",
    "Singleton",
    "NameSpace",
    "Str", "BaseStr",
    "List_", "BaseList",
    "Dict_", "DefaultDict", "BaseDict",
    "DateTime", "Date",
    "Frame",
    "Process",
    "Color",
    "Translator"
]

from .enums import Enum, ValueEnum, AutoEnum, ValueAutoEnum
from .markup import Html
from .http import Http
from .singleton import Singleton
from .namespace import NameSpace
from .translator import Translator
from .str import Str, BaseStr
from .list import List_, BaseList
from .dict import Dict_, DefaultDict, BaseDict
from .datetime import DateTime, Date
from .frame import Frame
from .process import Process
from .color import Color



from __future__ import annotations

import os
import subprocess
from typing import Any, Union, List

PathLike = Union[str, os.PathLike]


class CompletedProcess(subprocess.CompletedProcess):
    """A subclass of subprocess.CompletedProcess which can be tested for truthiness to return whether the process was successful."""

    def __bool__(self) -> bool:
        return self.returncode == 0


class Process(subprocess.Popen):
    """
    A subclass of subprocess.Popen which automatically sets up stdout and stderr pipes, and returns utf-8 decoded str rather than bytes from its streams.
    Can print the args passed to it when starting, and has a modified Process.wait() method that prints stdout in real time.
    """

    def __init__(self, args: List[str], cwd: PathLike = None, shell: bool = False, print_call: bool = True, stdout: Any = subprocess.PIPE, stderr: Any = subprocess.STDOUT, encoding: str = "utf-8", errors: str = "replace", **kwargs: Any) -> None:
        if cwd is not None and not shell:
            raise RuntimeError("'cwd' argument not supported without 'shell=True'")

        self.args = [str(arg) for arg in args]
        if print_call:
            print(self)

        super().__init__(args=self.args, stdout=stdout, stderr=stderr, shell=shell, cwd=cwd, encoding=encoding, errors=errors, **kwargs)  # type: ignore

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(str(self))})"

    def __str__(self) -> str:
        return subprocess.list2cmdline(self.args)

    def wait(self, timeout: float = None) -> CompletedProcess:  # type: ignore
        """Wait for the process to complete. Returns CompletedProcess rather than a returncode, and prints the stdout to the console in realtime"""
        if self.stdout is None:
            return CompletedProcess(self.args, returncode=super().wait(timeout=timeout), stdout=None)
        else:
            stdout = []
            while self.poll() is None:
                line = self.stdout.readline()
                print(line, end="")
                stdout.append(line)

            return CompletedProcess(self.args, returncode=self.poll(), stdout="".join(stdout))
